/*
 * Tencent is pleased to support the open source community by making BK-CI 蓝鲸持续集成平台 available.
 *
 * Copyright (C) 2019 THL A29 Limited, a Tencent company.  All rights reserved.
 *
 * BK-CI 蓝鲸持续集成平台 is licensed under the MIT license.
 *
 * A copy of the MIT License is included in this file.
 *
 *
 * Terms of the MIT License:
 * ---------------------------------------------------
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT
 * LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN
 * NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package com.tencent.devops.environment.service

import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import com.fasterxml.jackson.module.kotlin.readValue
import com.tencent.devops.common.api.enums.AgentStatus
import com.tencent.devops.common.api.exception.ErrorCodeException
import com.tencent.devops.common.api.pojo.OS
import com.tencent.devops.common.api.util.HashUtil
import com.tencent.devops.common.api.util.timestamp
import com.tencent.devops.common.auth.api.AuthPermission
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_BUILD_2_DEPLOY_DENY
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_BUILD_CAN_NOT_ADD_SVR
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_DEPLOY_2_BUILD_DENY
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_DEPLOY_CAN_NOT_ADD_AGENT
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_NO_CREATE_PERMISSSION
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_NO_DEL_PERMISSSION
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_NO_EDIT_PERMISSSION
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_ENV_NO_VIEW_PERMISSSION
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_NODE_INSUFFICIENT_PERMISSIONS
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_NODE_NAME_DUPLICATE
import com.tencent.devops.environment.constant.EnvironmentMessageCode.ERROR_NODE_NOT_EXISTS
import com.tencent.devops.environment.dao.EnvDao
import com.tencent.devops.environment.dao.EnvNodeDao
import com.tencent.devops.environment.dao.NodeDao
import com.tencent.devops.environment.dao.thirdPartyAgent.ThirdPartyAgentDao
import com.tencent.devops.environment.permission.EnvironmentPermissionService
import com.tencent.devops.environment.pojo.EnvCreateInfo
import com.tencent.devops.environment.pojo.EnvUpdateInfo
import com.tencent.devops.environment.pojo.EnvWithNodeCount
import com.tencent.devops.environment.pojo.EnvWithPermission
import com.tencent.devops.environment.pojo.EnvironmentId
import com.tencent.devops.environment.pojo.NodeBaseInfo
import com.tencent.devops.environment.pojo.enums.EnvType
import com.tencent.devops.environment.pojo.enums.NodeStatus
import com.tencent.devops.environment.pojo.enums.NodeType
import com.tencent.devops.environment.service.node.EnvCreatorFactory
import com.tencent.devops.environment.service.slave.SlaveGatewayService
import com.tencent.devops.environment.utils.AgentStatusUtils.getAgentStatus
import com.tencent.devops.environment.utils.NodeStringIdUtils
import org.jooq.DSLContext
import org.jooq.impl.DSL
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.stereotype.Service

@Service
class EnvService @Autowired constructor(
    private val dslContext: DSLContext,
    private val envDao: EnvDao,
    private val nodeDao: NodeDao,
    private val envNodeDao: EnvNodeDao,
    private val thirdPartyAgentDao: ThirdPartyAgentDao,
    private val slaveGatewayService: SlaveGatewayService,
    private val environmentPermissionService: EnvironmentPermissionService
) : IEnvService {

    override fun checkName(projectId: String, envId: Long?, envName: String) {
        if (envDao.isNameExist(dslContext, projectId, envId, envName)) {
            throw ErrorCodeException(errorCode = ERROR_NODE_NAME_DUPLICATE, params = arrayOf(envName))
        }
    }

    override fun createEnvironment(userId: String, projectId: String, envCreateInfo: EnvCreateInfo): EnvironmentId {
        if (!environmentPermissionService.checkEnvPermission(userId, projectId, AuthPermission.CREATE)) {
            throw ErrorCodeException(errorCode = ERROR_ENV_NO_CREATE_PERMISSSION)
        }

        checkName(projectId, null, envCreateInfo.name)

        val envCreator = EnvCreatorFactory.load(envCreateInfo.source.name)
            ?: throw IllegalArgumentException("unsupported nodeSourceType ${envCreateInfo.source}")

        return envCreator.createEnv(projectId = projectId, userId = userId, envCreateInfo = envCreateInfo)
    }

    override fun updateEnvironment(userId: String, projectId: String, envHashId: String, envUpdateInfo: EnvUpdateInfo) {
        val envId = HashUtil.decodeIdToLong(envHashId)
        if (!environmentPermissionService.checkEnvPermission(userId, projectId, envId, AuthPermission.EDIT)) {
            throw ErrorCodeException(errorCode = ERROR_ENV_NO_EDIT_PERMISSSION)
        }
        checkName(projectId, envId, envUpdateInfo.name)

        val existEnv = envDao.get(dslContext, projectId, envId)
        if (existEnv.envType == EnvType.BUILD.name && envUpdateInfo.envType != EnvType.BUILD) {
            throw ErrorCodeException(errorCode = ERROR_ENV_BUILD_2_DEPLOY_DENY)
        }
        if (existEnv.envType != EnvType.BUILD.name && envUpdateInfo.envType == EnvType.BUILD) {
            throw ErrorCodeException(errorCode = ERROR_ENV_DEPLOY_2_BUILD_DENY)
        }

        dslContext.transaction { configuration ->
            val context = DSL.using(configuration)

            envDao.update(
                dslContext = context,
                envId = HashUtil.decodeIdToLong(envHashId),
                name = envUpdateInfo.name,
                desc = envUpdateInfo.desc,
                envType = envUpdateInfo.envType.name,
                envVars = ObjectMapper().writeValueAsString(envUpdateInfo.envVars)
            )

            if (existEnv.envName != envUpdateInfo.name) {
                environmentPermissionService.updateEnv(
                    userId = userId,
                    projectId = projectId,
                    envId = envId,
                    envName = envUpdateInfo.name
                )
            }
        }
    }

    override fun listEnvironment(userId: String, projectId: String): List<EnvWithPermission> {
        val envRecordList = envDao.list(dslContext, projectId)
        if (envRecordList.isEmpty()) {
            return listOf()
        }

        val permissionMap = environmentPermissionService.listEnvByPermissions(
            userId,
            projectId,
            setOf(AuthPermission.LIST, AuthPermission.EDIT, AuthPermission.DELETE)
        )
        val canListEnvIds = if (permissionMap.containsKey(AuthPermission.LIST)) {
            permissionMap[AuthPermission.LIST]?.map { HashUtil.decodeIdToLong(it) } ?: emptyList()
        } else {
            emptyList()
        }

        val canEditEnvIds = if (permissionMap.containsKey(AuthPermission.EDIT)) {
            permissionMap[AuthPermission.EDIT]?.map { HashUtil.decodeIdToLong(it) } ?: emptyList()
        } else {
            emptyList()
        }

        val canDeleteEnvIds = if (permissionMap.containsKey(AuthPermission.DELETE)) {
            permissionMap[AuthPermission.DELETE]?.map { HashUtil.decodeIdToLong(it) } ?: emptyList()
        } else {
            emptyList()
        }

        val validRecordList = envRecordList.filter { canListEnvIds.contains(it.envId) }
        if (validRecordList.isEmpty()) {
            return listOf()
        }

        val nodeCountMap = envNodeDao.batchCount(dslContext, projectId, envRecordList.map { it.envId })
            .associateBy({ it.value1() }, { it.value2() })
        return envRecordList.map {
            EnvWithPermission(
                envHashId = HashUtil.encodeLongId(it.envId),
                name = it.envName,
                desc = it.envDesc,
                envType = if (it.envType == EnvType.TEST.name) EnvType.DEV.name else it.envType, // 兼容性代码
                nodeCount = nodeCountMap[it.envId] ?: 0,
                envVars = jacksonObjectMapper().readValue(it.envVars),
                createdUser = it.createdUser,
                createdTime = it.createdTime.timestamp(),
                updatedUser = it.updatedUser,
                updatedTime = it.updatedTime.timestamp(),
                canEdit = canEditEnvIds.contains(it.envId),
                canDelete = canDeleteEnvIds.contains(it.envId),
                canUse = null
            )
        }
    }

    override fun listUsableServerEnvs(userId: String, projectId: String): List<EnvWithPermission> {
        val envRecordList = envDao.listServerEnv(dslContext, projectId)
        if (envRecordList.isEmpty()) {
            return listOf()
        }

        val canUseEnvIds = environmentPermissionService.listEnvByPermission(userId, projectId, AuthPermission.USE)

        val validRecordList = envRecordList.filter { canUseEnvIds.contains(it.envId) }
        if (validRecordList.isEmpty()) {
            return listOf()
        }

        val nodeCountMap = envNodeDao.batchCount(dslContext, projectId, validRecordList.map { it.envId })
            .associateBy({ it.value1() }, { it.value2() })
        return validRecordList.map {
            EnvWithPermission(
                envHashId = HashUtil.encodeLongId(it.envId),
                name = it.envName,
                desc = it.envDesc,
                envType = if (it.envType == EnvType.TEST.name) EnvType.DEV.name else it.envType, // 兼容性代码
                nodeCount = nodeCountMap[it.envId] ?: 0,
                envVars = jacksonObjectMapper().readValue(it.envVars),
                createdUser = it.createdUser,
                createdTime = it.createdTime.timestamp(),
                updatedUser = it.updatedUser,
                updatedTime = it.updatedTime.timestamp(),
                canEdit = null,
                canDelete = null,
                canUse = null
            )
        }
    }

    override fun listEnvironmentByType(userId: String, projectId: String, envType: EnvType): List<EnvWithNodeCount> {
        val envRecordList = envDao.listByType(dslContext, projectId, envType)
        if (envRecordList.isEmpty()) {
            return emptyList()
        }

        val canListEnvIds = environmentPermissionService.listEnvByPermission(userId, projectId, AuthPermission.LIST)
        val validRecordList = envRecordList.filter { canListEnvIds.contains(it.envId) }
        if (validRecordList.isEmpty()) {
            return emptyList()
        }

        return validRecordList.map {
            val nodeIds = envNodeDao.list(dslContext, projectId, listOf(it.envId)).map { node ->
                node.nodeId
            }.toSet()

            val normalNodeCount = if (nodeIds.isEmpty()) {
                0
            } else {
                nodeDao.countNodeByStatus(dslContext, projectId, nodeIds, NodeStatus.NORMAL)
            }

            val abnormalNodeCount = if (nodeIds.isEmpty()) {
                0
            } else {
                nodeIds.size - normalNodeCount
            }

            EnvWithNodeCount(
                envHashId = HashUtil.encodeLongId(it.envId),
                name = it.envName,
                normalNodeCount = normalNodeCount,
                abnormalNodeCount = abnormalNodeCount
            )
        }
    }

    override fun listBuildEnvs(userId: String, projectId: String, os: OS): List<EnvWithNodeCount> {
        val envRecordList = envDao.listByType(dslContext, projectId, EnvType.BUILD)
        if (envRecordList.isEmpty()) {
            return emptyList()
        }

        val canListEnvIds = environmentPermissionService.listEnvByPermission(userId, projectId, AuthPermission.LIST)
        val validRecordList = envRecordList.filter { canListEnvIds.contains(it.envId) }
        if (validRecordList.isEmpty()) {
            return emptyList()
        }

        return validRecordList.map {
            val nodeIds = envNodeDao.list(dslContext, projectId, listOf(it.envId)).map { node ->
                node.nodeId
            }.toSet()

            val normalNodeCount = if (nodeIds.isEmpty()) {
                0
            } else {
                thirdPartyAgentDao.countAgentByStatusAndOS(dslContext, projectId, nodeIds, AgentStatus.IMPORT_OK, os)
            }

            val abnormalNodeCount = if (nodeIds.isEmpty()) {
                0
            } else {
                thirdPartyAgentDao.countAgentByStatusAndOS(
                    dslContext = dslContext,
                    projectId = projectId,
                    nodeIds = nodeIds,
                    status = AgentStatus.IMPORT_EXCEPTION,
                    os = os
                )
            }

            EnvWithNodeCount(
                envHashId = HashUtil.encodeLongId(it.envId),
                name = it.envName,
                normalNodeCount = normalNodeCount,
                abnormalNodeCount = abnormalNodeCount
            )
        }
    }

    override fun getEnvironment(userId: String, projectId: String, envHashId: String): EnvWithPermission {
        val envId = HashUtil.decodeIdToLong(envHashId)
        if (!environmentPermissionService.checkEnvPermission(userId, projectId, envId, AuthPermission.VIEW)) {
            throw ErrorCodeException(errorCode = ERROR_ENV_NO_VIEW_PERMISSSION)
        }
        val env = envDao.get(dslContext, projectId, envId)
        val nodeCount = envNodeDao.count(dslContext, projectId, envId)
        return EnvWithPermission(
            envHashId = HashUtil.encodeLongId(env.envId),
            name = env.envName,
            desc = env.envDesc,
            envType = if (env.envType == EnvType.TEST.name) EnvType.DEV.name else env.envType, // 兼容性代码
            nodeCount = nodeCount,
            envVars = jacksonObjectMapper().readValue(env.envVars),
            createdUser = env.createdUser,
            createdTime = env.createdTime.timestamp(),
            updatedUser = env.updatedUser,
            updatedTime = env.updatedTime.timestamp(),
            canEdit = environmentPermissionService.checkEnvPermission(userId, projectId, envId, AuthPermission.EDIT),
            canDelete = environmentPermissionService.checkEnvPermission(
                userId,
                projectId,
                envId,
                AuthPermission.DELETE
            ),
            canUse = null
        )
    }

    override fun listRawEnvByHashIds(
        userId: String,
        projectId: String,
        envHashIds: List<String>
    ): List<EnvWithPermission> {
        val envRecords =
            envDao.listServerEnvByIds(dslContext, projectId, envHashIds.map { HashUtil.decodeIdToLong(it) })
        return envRecords.map {
            EnvWithPermission(
                envHashId = HashUtil.encodeLongId(it.envId),
                name = it.envName,
                desc = it.envDesc,
                envType = if (it.envType == EnvType.TEST.name) EnvType.DEV.name else it.envType, // 兼容性代码
                nodeCount = null,
                envVars = jacksonObjectMapper().readValue(it.envVars),
                createdUser = it.createdUser,
                createdTime = it.createdTime.timestamp(),
                updatedUser = it.updatedUser,
                updatedTime = it.updatedTime.timestamp(),
                canEdit = null,
                canDelete = null,
                canUse = null
            )
        }
    }

    override fun listRawEnvByEnvNames(
        userId: String,
        projectId: String,
        envNames: List<String>
    ): List<EnvWithPermission> {
        val envRecords = envDao.listServerEnvByEnvNames(dslContext, projectId, envNames)
        val canUseEnvIds = environmentPermissionService.listEnvByPermission(userId, projectId, AuthPermission.USE)

        return envRecords.map {
            EnvWithPermission(
                envHashId = HashUtil.encodeLongId(it.envId),
                name = it.envName,
                desc = it.envDesc,
                envType = if (it.envType == EnvType.TEST.name) EnvType.DEV.name else it.envType, // 兼容性代码
                nodeCount = null,
                envVars = jacksonObjectMapper().readValue(it.envVars),
                createdUser = it.createdUser,
                createdTime = it.createdTime.timestamp(),
                updatedUser = it.updatedUser,
                updatedTime = it.updatedTime.timestamp(),
                canEdit = null,
                canDelete = null,
                canUse = canUseEnvIds.contains(it.envId)
            )
        }
    }

    override fun deleteEnvironment(userId: String, projectId: String, envHashId: String) {
        val envId = HashUtil.decodeIdToLong(envHashId)
        envDao.getOrNull(dslContext, projectId, envId) ?: return
        if (!environmentPermissionService.checkEnvPermission(userId, projectId, envId, AuthPermission.DELETE)) {
            throw ErrorCodeException(errorCode = ERROR_ENV_NO_DEL_PERMISSSION)
        }

        dslContext.transaction { configuration ->
            val context = DSL.using(configuration)
            envDao.deleteEnv(context, envId)
            environmentPermissionService.deleteEnv(projectId, envId)
        }
    }

    override fun listRawServerNodeByEnvHashIds(
        userId: String,
        projectId: String,
        envHashIds: List<String>
    ): Map<String, List<NodeBaseInfo>> {
        val envIds = envHashIds.map { HashUtil.decodeIdToLong(it) }
        val envNodes = envNodeDao.list(dslContext, projectId, envIds)
        val nodeRecords = nodeDao.listServerNodesByIds(dslContext, projectId, envNodes.map { it.nodeId })
        val nodeBaseInfos = nodeRecords.map { NodeStringIdUtils.getNodeBaseInfo(it) }
        val hashIdToNodeBaseInfoMap = nodeBaseInfos.associateBy { it.nodeHashId }

        val resultMap = mutableMapOf<String, MutableList<NodeBaseInfo>>()
        envHashIds.forEach {
            resultMap[it] = mutableListOf()
        }
        envNodes.forEach {
            val envHashId = HashUtil.encodeLongId(it.envId)
            val nodeHashId = HashUtil.encodeLongId(it.nodeId)
            if (hashIdToNodeBaseInfoMap.containsKey(nodeHashId)) {
                resultMap[envHashId]!!.add(hashIdToNodeBaseInfoMap[nodeHashId]!!)
            }
        }
        return resultMap
    }

    override fun listAllEnvNodes(userId: String, projectId: String, envHashIds: List<String>): List<NodeBaseInfo> {
        val envIds = envHashIds.map { HashUtil.decodeIdToLong(it) }
        val canUseEnvIdList = environmentPermissionService.listEnvByPermission(userId, projectId, AuthPermission.USE)
        val invalidEnvIds = envIds.filterNot { canUseEnvIdList.contains(it) }
        if (invalidEnvIds.isNotEmpty()) {
            throw ErrorCodeException(
                errorCode = ERROR_NODE_INSUFFICIENT_PERMISSIONS,
                params = arrayOf(invalidEnvIds.joinToString(","))
            )
        }

        val envNodeRecordList = envNodeDao.list(dslContext, projectId, envIds)
        val nodeIds = envNodeRecordList.map { it.nodeId }.toSet()
        val nodeList = nodeDao.listByIds(dslContext, projectId, nodeIds)

        val thirdPartyAgentMap =
            thirdPartyAgentDao.getAgentsByNodeIds(dslContext, nodeIds, projectId).associateBy { it.nodeId }
        return nodeList.map {
            val thirdPartyAgent = thirdPartyAgentMap[it.nodeId]
            val gatewayShowName = if (thirdPartyAgent != null) {
                slaveGatewayService.getShowName(thirdPartyAgent.gateway)
            } else {
                ""
            }

            val nodeStringId = NodeStringIdUtils.getNodeStringId(it)
            NodeBaseInfo(
                nodeHashId = HashUtil.encodeLongId(it.nodeId),
                nodeId = nodeStringId,
                name = it.nodeName,
                ip = it.nodeIp,
                nodeStatus = NodeStatus.getStatusName(it.nodeStatus),
                agentStatus = getAgentStatus(it),
                nodeType = NodeType.getTypeName(it.nodeType),
                osName = it.osName,
                createdUser = it.createdUser,
                operator = it.operator,
                bakOperator = it.bakOperator,
                gateway = gatewayShowName,
                displayName = NodeStringIdUtils.getRefineDisplayName(nodeStringId, it.displayName)
            )
        }
    }

    override fun addEnvNodes(userId: String, projectId: String, envHashId: String, nodeHashIds: List<String>) {
        val envId = HashUtil.decodeIdToLong(envHashId)
        if (!environmentPermissionService.checkEnvPermission(userId, projectId, envId, AuthPermission.EDIT)) {
            throw ErrorCodeException(errorCode = ERROR_ENV_NO_EDIT_PERMISSSION)
        }

        val nodeLongIds = nodeHashIds.map { HashUtil.decodeIdToLong(it) }

        // 检查 node 权限
        val canUseNodeIds = environmentPermissionService.listNodeByPermission(userId, projectId, AuthPermission.USE)
        val unauthorizedNodeIds = nodeLongIds.filterNot { canUseNodeIds.contains(it) }
        if (unauthorizedNodeIds.isNotEmpty()) {
            throw ErrorCodeException(
                errorCode = ERROR_NODE_INSUFFICIENT_PERMISSIONS,
                params = arrayOf(unauthorizedNodeIds.joinToString(",") { HashUtil.encodeLongId(it) })
            )
        }

        val env = envDao.get(dslContext, projectId, envId)

        // 检查 node 是否存在
        val existNodes = nodeDao.listByIds(dslContext, projectId, nodeLongIds)
        val existNodeIds = existNodes.map { it.nodeId }.toSet()
        val notExistNodeIds = nodeLongIds.filterNot { existNodeIds.contains(it) }
        if (notExistNodeIds.isNotEmpty()) {
            throw ErrorCodeException(
                errorCode = ERROR_NODE_NOT_EXISTS,
                params = arrayOf(notExistNodeIds.joinToString(",") { HashUtil.encodeLongId(it) })
            )
        }

        // 过滤已在环境中的节点
        val existEnvNodeIds = envNodeDao.list(dslContext, projectId, listOf(envId)).map { it.nodeId }
        val toAddNodeIds = nodeLongIds.subtract(existEnvNodeIds)

        // 验证节点类型
        val existNodesMap = existNodes.associateBy { it.nodeId }
        val serverNodeTypes = listOf(NodeType.CMDB.name, NodeType.CC.name, NodeType.BCSVM.name)

        toAddNodeIds.forEach {
            if (env.envType == EnvType.BUILD.name && existNodesMap[it]?.nodeType in serverNodeTypes) {
                throw ErrorCodeException(
                    errorCode = ERROR_ENV_BUILD_CAN_NOT_ADD_SVR,
                    params = arrayOf(HashUtil.encodeLongId(it))
                )
            }
            if (env.envType != EnvType.BUILD.name && existNodesMap[it]?.nodeType !in serverNodeTypes) {
                throw ErrorCodeException(
                    errorCode = ERROR_ENV_DEPLOY_CAN_NOT_ADD_AGENT,
                    params = arrayOf(HashUtil.encodeLongId(it))
                )
            }
        }

        envNodeDao.batchStoreEnvNode(dslContext, toAddNodeIds.toList(), envId, projectId)
    }

    override fun deleteEnvNodes(userId: String, projectId: String, envHashId: String, nodeHashIds: List<String>) {
        val envId = HashUtil.decodeIdToLong(envHashId)
        if (!environmentPermissionService.checkEnvPermission(userId, projectId, envId, AuthPermission.EDIT)) {
            throw ErrorCodeException(errorCode = ERROR_ENV_NO_EDIT_PERMISSSION)
        }

        envNodeDao.batchDeleteEnvNode(
            dslContext = dslContext,
            projectId = projectId,
            envId = HashUtil.decodeIdToLong(envHashId),
            nodeIds = nodeHashIds.map { HashUtil.decodeIdToLong(it) })
    }
}